/**
 * 
 */
package org.processmining.models.aggregatedactivitiesperformancediagram;

import java.awt.geom.Rectangle2D;
import java.math.BigInteger;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.processmining.framework.util.Pair;
import org.processmining.models.fuzzyperformancediagram.FPD;

/**
 * @author arya
 * @email arya.adriansyah@gmail.com
 * @version May 11, 2009
 */
public class AAPD {
	// index for relation time
	public static final int BAR_START_TIME = 0;
	public static final int BAR_SERVICE_TIME = 1;
	public static final int BAR_WAITING_TIME = 2;
	public static final int BAR_INTERSECTION_TIME = 3;
	public static final int BAR_THROUGHPUT_TIME = 4;

	// array index for relation time
	public static final int MIN_VALUE = 0;
	public static final int MAX_VALUE = 1;
	public static final int SUM = 2;
	public static final int SUM_SQUARE = 3;

	// data about other AAPDFocusElements
	private Map<AAPDElement, Map<AAPDElement, Map<Integer, Pair<BigInteger[], Integer>>>> relationTime;
	private Map<AAPDElement, Map<AAPDElement, Integer>> focusElementsFrequency;
	private Map<AAPDElement, Pair<Double, Double>> focusElementsBoundaries;

	// conceptual data
	private Set<AAPDElement> aapdElements;
	private String label;

	@SuppressWarnings("unused")
	private AAPD() {
	}

	/**
	 * Default constructor
	 * 
	 * @param fpd
	 */
	public AAPD(FPD fpd) {
		setLabel("AAPD of " + fpd.getLabel());
		aapdElements = new HashSet<AAPDElement>();
	}

	/**
	 * Initiate variables which will hold performance information between an
	 * element and the other
	 */
	public void initiateCalculation() { // initiate relation with others
		relationTime = new HashMap<AAPDElement, Map<AAPDElement, Map<Integer, Pair<BigInteger[], Integer>>>>(
				aapdElements.size());
		focusElementsFrequency = new HashMap<AAPDElement, Map<AAPDElement, Integer>>(aapdElements.size());
		focusElementsBoundaries = new HashMap<AAPDElement, Pair<Double, Double>>(aapdElements.size());

		// initiate relation with other nodes
		for (AAPDElement baseFocusElement : aapdElements) {
			Map<AAPDElement, Map<Integer, Pair<BigInteger[], Integer>>> mapRelTime = new HashMap<AAPDElement, Map<Integer, Pair<BigInteger[], Integer>>>();
			Map<AAPDElement, Integer> mapFreq = new HashMap<AAPDElement, Integer>();

			for (AAPDElement focusElement : aapdElements) {
				// relation
				Map<Integer, Pair<BigInteger[], Integer>> tempRelationMap = new HashMap<Integer, Pair<BigInteger[], Integer>>(
						5);

				for (int i = 0; i < 5; i++) {
					BigInteger[] bigIntArray = new BigInteger[4];
					bigIntArray[MIN_VALUE] = BigInteger.valueOf(Long.MAX_VALUE);
					bigIntArray[MAX_VALUE] = BigInteger.valueOf(Long.MIN_VALUE);
					bigIntArray[SUM] = BigInteger.ZERO;
					bigIntArray[SUM_SQUARE] = BigInteger.ZERO;

					Pair<BigInteger[], Integer> newPair = new Pair<BigInteger[], Integer>(bigIntArray, 0);
					tempRelationMap.put(i, newPair);
				}
				mapRelTime.put(focusElement, tempRelationMap);

				// frequency
				mapFreq.put(focusElement, 0);
			}

			relationTime.put(baseFocusElement, mapRelTime);
			focusElementsFrequency.put(baseFocusElement, mapFreq);
			focusElementsBoundaries.put(baseFocusElement, new Pair<Double, Double>(0.0, 0.0));
		}
	}

	/**
	 * Append statistical measurement between baseElement to targetElement
	 * 
	 * @param baseElement
	 * @param targetElement
	 * @param relationType
	 * @param values
	 */
	public void appendRelationTime(AAPDElement baseElement, AAPDElement targetElement, int relationType,
			List<Long> values) {
		assert (relationTime.containsKey(baseElement));
		if (values != null) {
			if (relationTime.get(baseElement).get(targetElement).containsKey(relationType)) {
				int counter = 0;
				BigInteger sum = BigInteger.ZERO;
				BigInteger sumSquare = BigInteger.ZERO;
				BigInteger min = BigInteger.valueOf(Long.MAX_VALUE);
				BigInteger max = BigInteger.valueOf(Long.MIN_VALUE);
				for (Long newValue : values) {
					BigInteger newValueInBigInt = BigInteger.valueOf(newValue);
					sum = sum.add(newValueInBigInt);
					sumSquare = sumSquare.add(newValueInBigInt.multiply(newValueInBigInt));

					if (newValueInBigInt.compareTo(min) == -1) {
						min = newValueInBigInt;
					}

					if (newValueInBigInt.compareTo(max) == 1) {
						max = newValueInBigInt;
					}
					counter++;
				}
				Pair<BigInteger[], Integer> oldPair = relationTime.get(baseElement).get(targetElement)
						.get(relationType);
				BigInteger[] oldLongArray = oldPair.getFirst();

				// update
				BigInteger[] newLongArray = new BigInteger[4];
				newLongArray[MIN_VALUE] = (oldLongArray[MIN_VALUE].compareTo(min) == -1) ? oldLongArray[MIN_VALUE]
						: min;
				newLongArray[MAX_VALUE] = (oldLongArray[MAX_VALUE].compareTo(max) == 1) ? oldLongArray[MAX_VALUE] : max;
				newLongArray[SUM] = oldLongArray[SUM].add(sum);
				newLongArray[SUM_SQUARE] = oldLongArray[SUM_SQUARE].add(sumSquare);

				Pair<BigInteger[], Integer> newPair = new Pair<BigInteger[], Integer>(newLongArray, oldPair.getSecond()
						+ counter);
				relationTime.get(baseElement).get(targetElement).put(relationType, newPair);
			} else {
				// this relation type is still new
				int counter = 0;
				BigInteger sum = BigInteger.ZERO;
				BigInteger sumSquare = BigInteger.ZERO;
				BigInteger min = BigInteger.valueOf(Long.MAX_VALUE);
				BigInteger max = BigInteger.valueOf(Long.MIN_VALUE);
				for (Long newValue : values) {
					BigInteger newValueInBigInt = BigInteger.valueOf(newValue);
					sum = sum.add(newValueInBigInt);
					sumSquare = sumSquare.add(newValueInBigInt.multiply(newValueInBigInt));

					if (newValueInBigInt.compareTo(min) == -1) {
						min = newValueInBigInt;
					}

					if (newValueInBigInt.compareTo(max) == 1) {
						max = newValueInBigInt;
					}
					counter++;
				}

				// update
				BigInteger[] newLongArray = new BigInteger[4];
				newLongArray[MIN_VALUE] = min;
				newLongArray[MAX_VALUE] = max;
				newLongArray[SUM] = sum;
				newLongArray[SUM_SQUARE] = sumSquare;

				Pair<BigInteger[], Integer> newPair = new Pair<BigInteger[], Integer>(newLongArray, counter);
				relationTime.get(baseElement).get(targetElement).put(relationType, newPair);
			}
		}
	}

	/**
	 * Add an element to an AAPD
	 * 
	 * @param label
	 * @return
	 */
	public AAPDElement addElement(String label) {
		AAPDElement bar = new AAPDElement(label);
		aapdElements.add(bar);
		return bar;
	}

	/**
	 * Set both lower and higher performance boundary for a baseElement
	 * 
	 * @param baseElement
	 * @param boundary
	 */
	public void setBoundaries(AAPDElement baseElement, Pair<Double, Double> boundary) {
		focusElementsBoundaries.put(baseElement, boundary);
	}

	/**
	 * Get lower and higher performance boundary of a baseElement
	 * 
	 * @param baseElement
	 * @return
	 */
	public Pair<Double, Double> getBoundaries(AAPDElement baseElement) {
		return focusElementsBoundaries.get(baseElement);
	}

	/**
	 * @return the relationTime
	 */
	public Map<AAPDElement, Map<AAPDElement, Map<Integer, Pair<BigInteger[], Integer>>>> getRelationTime() {
		return relationTime;
	}

	/**
	 * @param relationTime
	 *            the relationTime to set
	 */
	public void setRelationTime(
			Map<AAPDElement, Map<AAPDElement, Map<Integer, Pair<BigInteger[], Integer>>>> relationTime) {
		this.relationTime = relationTime;
	}

	/**
	 * @return the otherNodeCaseFrequency
	 */
	public Map<AAPDElement, Map<AAPDElement, Integer>> getFocusElementsCaseFrequency() {
		return focusElementsFrequency;
	}

	/**
	 * @param otherNodeCaseFrequency
	 *            the otherNodeCaseFrequency to set
	 */
	public void setFocusElementsCaseFrequency(Map<AAPDElement, Map<AAPDElement, Integer>> otherNodeCaseFrequency) {
		focusElementsFrequency = otherNodeCaseFrequency;
	}

	/**
	 * @return the aapdFocusElements
	 */
	public Set<AAPDElement> getAapdFocusElements() {
		return aapdElements;
	}

	/**
	 * @param aapdFocusElements
	 *            the aapdFocusElements to set
	 */
	public void setAapdFocusElements(Set<AAPDElement> aapdFocusElements) {
		aapdElements = aapdFocusElements;
	}

	/**
	 * @return the label
	 */
	public String getLabel() {
		return label;
	}

	/**
	 * @param label
	 *            the label to set
	 */
	public void setLabel(String label) {
		this.label = label;
	}

	/**
	 * Increment case frequency in which baseFocusElement and targetFocusElement
	 * occurs. This method needs to be executed for both baseElement and
	 * targetElement to update values for both
	 * 
	 * @param baseElement
	 * @param targetElement
	 */
	public void incrementFocusElementsCaseFrequency(AAPDElement baseElement, AAPDElement targetElement) {
		assert (focusElementsFrequency.get(baseElement).containsKey(targetElement));
		focusElementsFrequency.get(baseElement).put(targetElement,
				focusElementsFrequency.get(baseElement).get(targetElement) + 1);
	}

	/**
	 * get the bound of an AAPD
	 * 
	 * @return
	 */
	public Rectangle2D getBounds() {
		double x = Double.MAX_VALUE, y = Double.MAX_VALUE, w = 0, h = 0;
		return new Rectangle2D.Double(x, y, w - x, h - y);
	}

}
